import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
import bz2
import requests
import os.path as path
import pickle
import reportlab as rl
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Image
from reportlab.lib.units import inch
from reportlab.lib.styles import ParagraphStyle
import glob
import itertools
from PIL import Image as PIL_Image
import random
from sklearn.decomposition import PCA
import warnings
from sklearn.manifold import TSNE


# ----------------------------------------------------------------------------------- #
#                                      Decorators                                     #
# ----------------------------------------------------------------------------------- #


def ignore_warnings(argument):
    def _ignore_warnings(function):
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", argument)
                return function(*args, **kwargs)
        return wrapper
    return _ignore_warnings


# ----------------------------------------------------------------------------------- #
#                                   KL-divergences                                    #
# ----------------------------------------------------------------------------------- #


def _mc_integral_kl(st_a, st_b, n_samp):
    samples = st_b.sample(sample_shape=n_samp)
    log_density_ratio = st_a.log_prob(samples) - st_b.log_prob(samples)
    density_a = st_a.prob(samples)
    return tf.reduce_mean(density_a * log_density_ratio, axis=0)


# TODO only register these if they aren't already so this module can be reloaded
# this is from here: https://mast.queensu.ca/~communications/Papers/GiAlLi13.pdf
@tf.distributions.RegisterKL(tf.distributions.Laplace, tf.distributions.Laplace)
def _kl_laplace(lap_i, lap_j, name=None):
    return tf.log(lap_j.scale / lap_i.scale) + tf.abs(lap_i.loc - lap_j.loc) / lap_j.scale + (lap_i.scale / lap_j.scale) * tf.exp(-tf.abs(lap_i.loc - lap_j.loc) / lap_i.scale) - 1


# monte carlo integral, could use a closed-form formula instead
@tf.distributions.RegisterKL(tf.distributions.StudentT, tf.distributions.StudentT)
def _kl_student(st_a, st_b, name=None):
    return _mc_integral_kl(st_a, st_b, 1)


# TODO there should be a closed-form solution for this
@tf.distributions.RegisterKL(tf.distributions.Normal, tf.distributions.Laplace)
def _kl_gauss_laplace(st_a, st_b, name=None):
    return _mc_integral_kl(st_a, st_b, 1)


# ----------------------------------------------------------------------------------- #
#                                   Fetching datasets                                 #
# ----------------------------------------------------------------------------------- #


def loadImages(nPX, samplesPerLabel, kind='textures', folder='/home/banmi/imageDB/textureSynth/bin/', use_labels=None):
    fname = '%s_%d_%dpx.pkl' % (kind, samplesPerLabel, nPX)
    if not path.isfile(folder + fname):
        # if the data is not available locally, dowload it
        print("Downloading data...")
        r = requests.get("http://www.rmki.kfki.hu/~banmi/textures/" + fname + '.bz2')
        print("Decompressing data...")
        bin_data = bz2.decompress(r.content)
        with open(fname, 'wb') as imgfile:
            imgfile.write(bin_data)
        folder = './'
    
    with open(folder + fname, 'rb') as file:
        textures = pickle.load(file)
    if isinstance(textures, dict) and use_labels is not None:
        use_indices_train = [i in use_labels for i in textures['train_labels']]
        use_indices_test = [i in use_labels for i in textures['test_labels']]
        textures['train_images'] = textures['train_images'][use_indices_train, :]
        textures['train_labels'] = textures['train_labels'][use_indices_train]
        textures['test_images'] = textures['test_images'][use_indices_test, :]
        textures['test_labels'] = textures['test_labels'][use_indices_test]
    return textures


# ----------------------------------------------------------------------------------- #
#                                Gabor filter creation                                #
# ----------------------------------------------------------------------------------- #


def gabor_patch(size, lambda_, theta, sigma, phase, trim=.005):
    """
    size : int Image size (n x n)
    lambda_ : int Spatial frequency (px per cycle)
    theta : int or float Grating orientation in degrees
    sigma : int or float gaussian standard deviation (in pixels)
    phase : float 0 to 1 inclusive
    """
    # make linear ramp
    X0 = (np.linspace(1, size, size) / size) - .5
    # Set wavelength and phase
    freq = size / float(lambda_)
    phaseRad = phase * 2 * np.pi
    # Make 2D grating
    Xm, Ym = np.meshgrid(X0, X0)
    # Change orientation by adding Xm and Ym together in different proportions
    thetaRad = (theta / 360.) * 2 * np.pi
    Xt = Xm * np.cos(thetaRad)
    Yt = Ym * np.sin(thetaRad)
    grating = np.sin(((Xt + Yt) * freq * 2 * np.pi) + phaseRad)
    # 2D Gaussian distribution
    gauss = np.exp(-((Xm ** 2) + (Ym ** 2)) / (2 * (sigma / float(size)) ** 2))
    # Trim
    gauss[gauss < trim] = 0
    return grating * gauss


def roll_zeropad(a, shift, axis=None):
    a = np.asanyarray(a)
    if shift == 0:
        return a
    if axis is None:
        n = a.size
        reshape = True
    else:
        n = a.shape[axis]
        reshape = False
    if np.abs(shift) > n:
        res = np.zeros_like(a)
    elif shift < 0:
        shift += n
        zeros = np.zeros_like(a.take(np.arange(n - shift), axis))
        res = np.concatenate((a.take(np.arange(n - shift, n), axis), zeros), axis)
    else:
        zeros = np.zeros_like(a.take(np.arange(n - shift, n), axis))
        res = np.concatenate((zeros, a.take(np.arange(n - shift), axis)), axis)
    if reshape:
        return res.reshape(a.shape)
    else:
        return res


def gabor_filter_bank(nFilter, nPX, wavelengths):
    filters = np.zeros((nFilter, nPX, nPX))
    unshifted = np.zeros((nFilter, nPX, nPX))
    for f in range(nFilter):
        position_px = np.floor(nPX * np.random.uniform(size=2)).astype(int)
        shift_px = (position_px - np.floor(nPX / 2)).astype(int)

        orientation_deg = np.random.randint(180)
        wavelength_px = wavelengths[np.random.randint(len(wavelengths))]  # TODO use real distribution
        SD_px = np.floor(wavelength_px / 1.2)  # TODO make elongated ellipses
        phase = 0  # TODO do we need to vary this?
        actFilter = gabor_patch(nPX, wavelength_px, orientation_deg, SD_px, phase)
        unshifted[f, :, :] = actFilter
        filters[f, :, :] = roll_zeropad(roll_zeropad(actFilter, shift_px[0], axis=0), shift_px[1], axis=1)
    return filters


# ----------------------------------------------------------------------------------- #
#                                   Image generation                                  #
# ----------------------------------------------------------------------------------- #


def get_white_noise_image(width, height):
    pil_map = PIL_Image.new("L", (width, height), 255)
    random_grid = map(lambda x: (
            int(random.random() * 256)
        ), [0] * width * height)
    pil_map.putdata(list(random_grid))
    return np.array(pil_map)


def cropImages(imgs, crop_size):
    n_img = imgs.shape[0]
    orig_px = int(np.sqrt(imgs.shape[1]))
    return imgs.reshape(n_img, orig_px, orig_px)[:, :crop_size, :crop_size].reshape(n_img, crop_size * crop_size)


def whitenImages(imgs, filter_noise=True, pca=None):
    if pca is None:
        pca = PCA()
        pca.fit(imgs)
    latents = pca.transform(imgs)
    normed_s = 1.0 / np.sqrt(pca.explained_variance_)
    if filter_noise:
        n_retained_comps = int(pca.explained_variance_.shape[0] * np.pi / 4.0)
        normed_s[n_retained_comps:] = 0
    return np.matmul(latents, np.matmul(np.diag(normed_s), pca.components_)), pca


def matchPixelStats(images, sdMatch, contrastTuningExp=1):
    matchedImages = np.zeros((0, images.shape[1]))
    for image in images:
        image -= np.mean(image)
        image = (sdMatch / np.std(image)) ** contrastTuningExp * (image - np.mean(image)) + np.mean(image)    
        matchedImages = np.vstack([matchedImages, image])
    return matchedImages


# ----------------------------------------------------------------------------------- #
#                                Correlation matrices                                 #
# ----------------------------------------------------------------------------------- #


def corrMatAbsDiffVals(M1, M2):
    return np.abs(M1 - M2)[np.triu_indices(M1.shape[0], 1)]


# ----------------------------------------------------------------------------------- #
#                                       Plotting                                      #
# ----------------------------------------------------------------------------------- #


def grey_image(img, unitscale=True, ylabel=None, fontsize=26, border_color=None):
    if img.ndim == 1:
        nPX = int(np.sqrt(img.shape[0]))
        img = img.reshape(nPX, nPX)

    if not isinstance(unitscale, bool):
        vlimits = {'vmin': -unitscale, 'vmax': unitscale}
    elif unitscale:
        vlimits = {'vmin': 0, 'vmax': 1}
    else:
        vlimits = {}

    plt.imshow(img, interpolation='none', cmap=plt.get_cmap('gray'), **vlimits)

    if border_color is not None:
        for spine in ['bottom', 'top', 'left', 'right']:
            plt.gca().spines[spine].set_color(border_color)
            plt.gca().spines[spine].set_linewidth(4)
    plt.gca().set_xticks([])
    plt.gca().set_yticks([])
    if ylabel is not None:
        plt.ylabel(ylabel, fontsize=fontsize)


def image_set(images, plotter_func=grey_image, samplePerRow=20, unitscale=False, file_name=None):    
    samplePerRow = np.minimum(samplePerRow, images.shape[0])
    nRows = np.ceil(images.shape[0] / samplePerRow)
    f = plt.figure(figsize=(3 * samplePerRow, 3 * nRows))
    for i in range(images.shape[0]):
        plt.subplot(nRows, samplePerRow, i + 1)
        grey_image(images[i, ], unitscale)
    if file_name is not None:
            f.savefig(file_name, bbox_inches='tight')


def doubleCorrPlot(M1, M2, max_val=None):
    m = np.triu(M1) + np.tril(M2) - 2 * np.eye(M1.shape[0])
    plt.imshow(m, cmap='coolwarm', vmin=-1 if max_val is None else -max_val, vmax=1 if max_val is None else max_val)
    plt.gca().set_xticks([])
    plt.gca().set_yticks([])


def tsne_plot(data, labels=None, tsne_perp=20, tsne_ndata=2000):
    random_indices = np.random.randint(0, data.shape[0] - 1, size=tsne_ndata)
    model = TSNE(n_components=2, perplexity=tsne_perp, verbose=False)
    np.set_printoptions(suppress=True)
    tsne_data = model.fit_transform(data[(random_indices)])
    plt.scatter(tsne_data[:, 0], tsne_data[:, 1], c=np.array(labels)[(random_indices)] if labels is not None else 'b', s=15, cmap='Set1')
    plt.title('t-SNE perp %d ndata %d' % (tsne_perp, tsne_ndata), fontsize=12)


@ignore_warnings(ResourceWarning)
def createReport(folder_prefix, imgfile_postfix, img_height_in, img_width_in=6, base_folder='results/saved/', single_folder=False):
    folders = glob.glob(base_folder + folder_prefix + ('*' if not single_folder else '/*' + imgfile_postfix))
    names = [f.split(folder_prefix)[-1] for f in folders] if not single_folder else folders
    names.sort()

    doc = SimpleDocTemplate('results/reports/REPORT_' + folder_prefix + '_' + imgfile_postfix.split('.')[0] + '.pdf')
    pstyle = ParagraphStyle('e')
    titles = [Paragraph(s, pstyle) for s in names]
    images = [Image((base_folder + folder_prefix + s + '/' + folder_prefix + s + imgfile_postfix if not single_folder else s) , 
                    img_width_in * inch, img_height_in * inch) for s in names] 
    doc.build(list(itertools.chain(*zip(titles, images))))
