import tensorflow as tf
import warnings
# tensorflow_probability is compiled with an earlier numpy version and it throws a stupid warning so we have to import it in a stupid way
with warnings.catch_warnings():
    warnings.simplefilter("ignore", RuntimeWarning)
    from tensorflow_probability import distributions as tfd
import matplotlib.pyplot as plt
import numpy as np
import pickle
import random
import pandas as pd
import seaborn as sns
import helperfun as hf
from abc import ABC
from abc import abstractmethod
from sklearn.manifold import TSNE
import tensorflow_probability as tfp


# ----------------------------------------------------------------------------------- #
#                                  LAYER FUNCTIONS                                    #
# ----------------------------------------------------------------------------------- #


def dense_layer(inputs, units, activation, init_sd):
    # this basically just makes the calls shorter
    return tf.layers.dense(inputs=inputs, units=units, activation=activation,
                           bias_initializer=tf.constant_initializer(init_sd),
                           kernel_initializer=tf.truncated_normal_initializer(stddev=init_sd))


def conv_layer(inputs, filters, kernel_size=5, strides=2, activation=tf.nn.relu, flatten=False, keep_prob=0.8, transpose=False):
    conv_params = {'inputs': inputs, 'filters': filters, 'kernel_size': kernel_size, 'strides': strides, 'padding': 'same', 'activation': activation}
    layer_function = tf.layers.conv2d if not transpose else tf.layers.conv2d_transpose
    return tf.nn.dropout(layer_function(**conv_params), keep_prob)


# ----------------------------------------------------------------------------------- #
#                            ENCODERS AND DECODERS                                    #
# ----------------------------------------------------------------------------------- #


class Coder(ABC):
    # this is an abstract base class for all encoders and decoders
    output_layer = True

    @abstractmethod
    def build(self, input_layer):
        # it should give back the reference to the last layer
        # assumes that the default graph is where we want to put this
        pass


class DenseCoder(Coder):
    def __init__(self, nLayer=2, units=500, activation=tf.nn.sigmoid, init_sd=0.1):
        self.units = units
        self.activation = activation
        self.init_sd = init_sd
        self.nLayer = nLayer

    def build(self, input_layer):
        layers = [dense_layer(inputs=input_layer, units=self.units, activation=self.activation, init_sd=self.init_sd)]
        for l in range(1, self.nLayer):
            layers.append(dense_layer(inputs=layers[-1], units=self.units, activation=self.activation, init_sd=self.init_sd))
        return layers[-1]


class LinearCoder(DenseCoder):
    output_layer = False

    def __init__(self, units):
        super().__init__(nLayer=1, units=units, activation=None)


class ConvolutionalCoder(Coder):
    def __init__(self, nLayer=3, nFilter=[64, 64, 64], kernel_size=5, activation=tf.nn.relu, strides=[2, 2, 1], keep_prob=0.8, transpose_convolution=False):
        assert len(strides) == nLayer, "Please provide %d stride values." % nLayer
        assert len(nFilter) == nLayer, "Please provide %d filter values." % nLayer
        self.nLayer = nLayer
        self.nFilter = nFilter
        self.kernel_size = kernel_size
        self.activation = activation
        self.strides = strides
        self.keep_prob = keep_prob
        self.transpose_convolution = transpose_convolution  # setting this to True creates a 'deconv' coder

    def build(self, input_layer):
        assert np.sqrt(int(input_layer.shape[1])).is_integer(), 'This VAE only works with latent dimensions that are square numbers.'
        in_dim = np.sqrt(int(input_layer.shape[1])).astype(int)
        reshaped_input = tf.reshape(input_layer, shape=[-1, in_dim, in_dim, 1])
        layers = [reshaped_input]
        for l in range(1, self.nLayer):
            layers.append(conv_layer(inputs=layers[-1],
                                     filters=self.nFilter[l - 1], kernel_size=self.kernel_size, strides=self.strides[l - 1],
                                     activation=self.activation, keep_prob=self.keep_prob,
                                     transpose=self.transpose_convolution))
        return tf.contrib.layers.flatten(layers[-1])


class GaborDecoder(Coder):
    output_layer = False  # this decoder generates the Bernoulli parameters themselves, no need for an extra dense layer

    def __init__(self, data_dim, nFilter, wavelengths=[2, 4], init_sd=0.1, connectingLayer=False, plotFilters=False):
        self.nFilter = nFilter
        self.data_dim = data_dim
        self.wavelengths = wavelengths
        self.init_sd = init_sd
        self.plotFilters = plotFilters
        self.connectingLayer = connectingLayer

    def build(self, input_layer):
        nPX = int(np.sqrt(self.data_dim))
        filters = np.float32(hf.gabor_filter_bank(self.nFilter, nPX, self.wavelengths).reshape(self.nFilter, nPX * nPX))
        if self.plotFilters:
            hf.image_set(filters[1:20, :], unitscale=False)
        filter_act = dense_layer(inputs=input_layer, units=self.nFilter, activation=None, init_sd=self.init_sd) if self.connectingLayer else input_layer
        return tf.matmul(filter_act, filters)


# ----------------------------------------------------------------------------------- #
#                                   MAIN VAE CLASS                                    #
# ----------------------------------------------------------------------------------- #

class VAE():
    """Handle the definition, training, inference and diagnostics of a variational autoencoder.

    Initializer args:
        - name (str): used when saving and loading weights.
        - seed (int): you can control the state of the random number generator by supplying this

        Model dimensions
        - data_dim (int): the number of observed dimensions (that of x)
        - batch_size (int): number of data points considered in one inference/gradient step
        - latent_dim (int): number of stochastic latent units in the VAE (usually called z). Must be a square number.

        Model architecture
        - encoder (vae.Coder): the recognitive model / encoder mapping from x to z. Its output is mapped to the parameters of the posterior p(z | x) by linear dense layers.
        - decoder (vae.Coder): the generative model / decoder mapping from z to the parameters / a sample from p(x | z). Its output is mapped to the parameters of
          the likelihood p(x | z) by linear dense layers unless the object's output_layer property is False, in which case it is required to have the same dimensionality on
          the output as the likelihood parameters.

        Model distributions
        - prior_distr (tensorflow_probability.distributions.Distribution): functional form of both the prior p(z) and the posterior p(z | x) distributions.
          Now accepts Dirichlet, MultivariateNormalTriL or anything that is a location-scale family distribution (tested with Normal and Laplace).
        - prior_scale (float): how much is the prior concentrated. In case of Dirichlet priors, in the corners of the simplex, in case of location-scale priors, around 0 (the MVN prior is also independent, only the posterior allows for correlations).
        - likelihood (str): functional form of the likelihood p(x | z) distribution. Now accepts either 'Gauss' or 'Bernoulli'.
        - observ_sd (float or str or None): only considered when the likelihood has a deviation/variance parameter (only in the case of the Gauss likelihood now).
          If None, the likelihood variance is mapped from z by the decoder, if 'learn_common', then the likelihood variance is a learned scalar parameter of the model,
          if a number, the likelihood variance is set to a constant of its value.

        Learning
        - cost_summary (function): how to map the reconstruction and KL costs computed for the vectors of observed and latent variables respectively to scalar values
          for each batch. Good choices are tensorflow.reduce_sum, which gives the exact log-likelihoods and KLs, and tensorflow.reduce_mean, that only differs in a constant,
          but makes calibrating beta much easier
        - beta (float): sets the balance between the reconstruction and KL terms in the cost function. Values less then 1 favor minimizing the reconstruction error.
        - learning_rate (float): for the gradient descent optimizer
        - init_stdev (float): all learnable parameters of the model are initialized stochastically using this standard deviation value.
    """
    
    @hf.ignore_warnings(UserWarning)  # opening more than one session throws a warning but generally does not pose a real problem
    def __init__(self, data_dim, batch_size, latent_dim, encoder, decoder, beta, learning_rate, likelihood='Bernoulli', gauss_posterior=False,
                 prior_distr=tfd.Normal, init_stdev=0.1, prior_scale=1.0, name='vae', observ_sd=None, seed=None, cost_summary=tf.reduce_mean):
        assert np.sqrt(data_dim).is_integer(), 'This VAE only works with observed dimensions that are square numbers.'
        self.name = name
        self.data_dim = data_dim
        self.nPX = int(np.sqrt(self.data_dim))
        self.batch_size = batch_size
        self.latent_dim = latent_dim
        self.beta = beta
        self.learning_rate = learning_rate
        self.prior_mu = 0.0 * np.ones((self.batch_size, self.latent_dim), dtype=np.float32)
        self.prior_std = prior_scale * np.ones((self.batch_size, self.latent_dim), dtype=np.float32)

        self.graph = tf.Graph()
        if seed is not None:
            tf.set_random_seed(seed)

        with self.graph.as_default():
            # Input
            self.x = tf.placeholder(tf.float32, shape=[None, self.data_dim], name='x')

            # Encoder
            self.encoder = encoder.build(self.x)

            # Posterior parameters - TODO rename them so we can see it's the posterior and not something else
            self.sigma_pre = dense_layer(inputs=self.encoder, units=self.latent_dim, activation=None, init_sd=init_stdev)

            if prior_distr == tfd.Dirichlet:
                self.alpha = tf.exp(self.sigma_pre)                
                self.mu = self.alpha  # this is needed because plotters expect mu to be here

                self.q = prior_distr(concentration=self.alpha)  # the posterior is always from the same family as the prior
                self.prior = prior_distr(concentration=self.prior_std)  # so that the prior_scale parameter acts on it
            elif prior_distr == tfd.MultivariateNormalTriL:
                self.q = tfp.trainable_distributions.multivariate_normal_tril(self.encoder, self.latent_dim)
                single_batch_scale = prior_scale * tf.eye(self.latent_dim)
                self.prior_scale_tril = tf.tile(tf.expand_dims(single_batch_scale, axis=0), [self.batch_size, 1, 1])  # TODO this might be slow
                self.prior = prior_distr(loc=self.prior_mu, scale_tril=self.prior_scale_tril)
                self.mu = self.q.loc                
                self.std = self.q.scale.diag_part()
            else:
                # when the posterior is not a Dirichlet, it is assumed to be a location-scale family, e.g. Gauss or Laplace
                self.std = tf.sqrt(tf.exp(self.sigma_pre))
                #self.std = tf.sigmoid(self.sigma_pre)
                self.mu = dense_layer(inputs=self.encoder, units=self.latent_dim, activation=None, init_sd=init_stdev)
                
                if not gauss_posterior:
                    self.q = prior_distr(loc=self.mu, scale=self.std)
                else:
                    # TODO enable this option for the non-location-scale priors too
                    self.q = tfd.Normal(loc=self.mu, scale=self.std)
                self.prior = prior_distr(loc=self.prior_mu, scale=self.prior_std)

            # Sampling the posterior
            self.z = self.q.sample()

            # Decoder
            self.decoder = decoder.build(self.z)

            # Reconstruction parameters - TODO they should be parametrized the same way as the prior and the posterior
            if likelihood == 'Bernoulli':
                if decoder.output_layer:
                    self.x_rec_mean = dense_layer(inputs=self.decoder, units=self.data_dim, activation=tf.nn.sigmoid, init_sd=init_stdev)
                else:
                    self.x_rec_mean = tf.sigmoid(self.decoder)
                self.x_rec_sd = tf.constant(0)
                # if you don't make it sufficiently far from both 0 and 1, nan is waiting for you
                self.observation = tfd.Bernoulli(probs=tf.minimum(self.x_rec_mean + 1e-10, 1.0 - 1e-6))

            elif likelihood == 'Gauss':
                if decoder.output_layer:
                    self.x_rec_mean = dense_layer(inputs=self.decoder, units=self.data_dim, activation=None, init_sd=init_stdev)
                    if observ_sd is None:
                        # TODO this should be set up in such a way that the SD does not end up very near 0
                        self.x_rec_sd = tf.sqrt(tf.exp(dense_layer(inputs=self.decoder, units=self.data_dim, activation=None, init_sd=init_stdev)))
                else:
                    self.x_rec_mean = self.decoder
                    if observ_sd is None:
                        observ_sd = 0.1  # TODO this is not thought through very well

                if observ_sd == 'learn_common':
                    self.observ_sd = tf.Variable(tf.random_uniform(shape=[1], minval=0.01, maxval=0.1))
                    self.x_rec_sd = self.observ_sd * tf.ones((self.batch_size, self.data_dim))
                elif observ_sd is not None:
                    self.x_rec_sd = tf.constant(observ_sd * np.ones((self.batch_size, self.data_dim), dtype=np.float32))

                self.observation = tfd.Normal(loc=self.x_rec_mean, scale=self.x_rec_sd)
            else:
                raise ValueError("Unknown likelihood function.")

            self.full_rec = self.observation.log_prob(self.x)
            self.reconstruction_cost = - cost_summary(self.full_rec, axis=1)

            self.full_kl = tfd.kl_divergence(self.q, self.prior)            
            # self.kl_cost = cost_summary(self.full_kl, axis=1) if not prior_distr == tfd.Dirichlet else self.full_kl
            self.kl_cost = cost_summary(self.full_kl, axis=1) if not self.full_kl.shape == (self.batch_size, ) else self.full_kl

            self.actBeta = tf.placeholder(tf.float32, name='beta')

            self.cost = tf.reduce_mean(self.reconstruction_cost + self.actBeta * self.kl_cost)

            # Optimizer
            self.optimizer = tf.train.AdamOptimizer(learning_rate=self.learning_rate).minimize(self.cost)

            # Sample from the likelihood, if we ever need it
            self.likelihood_sample = self.observation.sample()

            # Init
            self.training_epochs = 0
            self.saver = tf.train.Saver()
            self.sess = tf.InteractiveSession()  # this way a session is opened at every instantiation of a VAE, generating warnings but not posing a problems otherwise
            self.init = tf.global_variables_initializer()
            self.sess.run(self.init)

    def _get_batch(self, data):
        # TODO do this with a generator
        idx = random.sample(range(data.shape[0]), self.batch_size)
        return data[idx, :], idx

    def train(self, train_data, runs, reinit=False, verbose=False, print_cycle=10, save_weights=None, burnin_epochs=0, burnin_mode='anneal_beta', goal_beta=None):
        nData = train_data.shape[0]
        n_minibatches = int(nData / self.batch_size)  # TODO handle non-integer cases
        
        if goal_beta is not None:
            self.beta = goal_beta
        
        with self.graph.as_default():
            if reinit:
                self.training_epochs = 0
                self.sess.run(self.init)
            for epoch in range(runs):
                if verbose:
                    pbar = tf.contrib.keras.utils.Progbar(n_minibatches)

                # set beta
                actBeta = self.beta
                if epoch < burnin_epochs:
                    if burnin_mode == 'hold_beta':
                        actBeta = 0.000000001
                    elif burnin_mode == 'anneal_beta':
                        actBeta = self.beta * (epoch + 1) / burnin_epochs  # TODO figure out how this works best

                for i in range(n_minibatches):
                    x_batch, idx = self._get_batch(train_data)

                    # some diagnostics
                    if epoch == 0 and i == 0:
                        kl_, rec_, x_m, x_s, f_r, post_mu, sd_pre, full_kl = self.sess.run((self.kl_cost, self.reconstruction_cost, self.x_rec_mean, self.x_rec_sd, 
                                                                                    self.full_rec, self.mu, self.sigma_pre, self.full_kl), feed_dict={self.x: x_batch})
                        print("Initial rec: %f, kl: %f, about %f times the reconstruction term." % (np.mean(rec_), np.mean(kl_), np.mean(kl_ / rec_)))
                        # print(full_kl[0, :10])                        
                        # print(post_sd[0, :10])
                        # print(sd_pre)
                        if np.isnan(np.sum(kl_)) or not np.all(np.isfinite(kl_)):
                            print(sd_pre)
                            raise FloatingPointError('The rec error came back nan in %d places out of %d.' % (np.sum(np.isnan(f_r)), self.data_dim * self.batch_size))
                        if np.isnan(np.sum(rec_)):
                            print(x_m[np.isnan(f_r)])
                            raise FloatingPointError('The rec error came back nan in %d places out of %d.' % (np.sum(np.isnan(f_r)), self.data_dim * self.batch_size))
                        elif np.mean(rec_) > 1e10:
                            if x_s is not None:
                                print(x_s[0, f_r[0, :] < -1e10])
                            raise FloatingPointError('The rec error is gigantic, you probably have too small likelihood variances for some dimensions.')

                    # main optimization step
                    _, cost_, kl_, rec_, mu_, xsd_, enc_ = self.sess.run((self.optimizer, self.cost, self.kl_cost, self.reconstruction_cost, self.mu, self.x_rec_sd, self.encoder), feed_dict={self.x: x_batch, self.actBeta: actBeta})
                    if np.isnan(np.sum(kl_)):
                        print(mu_)
                        print(enc_)
                        raise FloatingPointError('The KL came back nan.')
                    if np.isnan(np.sum(rec_)):
                        print(xsd_)
                        raise FloatingPointError('The reconstruction error came back nan.')

                    if verbose:
                        pbar.add(1, [("avg cost", cost_), ("kl cost", kl_), ("rec cost", rec_)])
                if verbose:
                    print("\n")
                elif (epoch + 1) % print_cycle == 0:
                    print("Epoch %d, rec %f, kl %f, cost %f, beta %f " % (epoch + 1, np.mean(rec_), np.mean(kl_), np.mean(cost_), actBeta))
                    # self.plotReconstruction(test_data=train_data, n_rec=1, simplified_output=True)
            if verbose:
                print("final cost:", cost_)
            self.training_epochs += runs
            if save_weights is not None:
                # TODO check if folder exists and create it if it doesn't
                self.saver.save(self.sess, save_weights + self.name + '_%dksamp_%depoch.ckpt' % (int(train_data.shape[0] / 1000), self.training_epochs), write_meta_graph=False)

    def sampleLatents(self, data, labels=None, file_name=None):
        nData = data.shape[0]
        n_minibatches = int(nData / self.batch_size)
        zs = {'map': np.zeros((0, self.latent_dim)), 'samples': np.zeros((0, self.latent_dim))}
        if labels is not None:
            zs['labels'] = np.zeros(0)
        if hasattr(self, 'std'):
            zs['std'] = np.zeros((0, self.latent_dim))
        with self.graph.as_default():
            for i in range(n_minibatches):
                x_batch, idx = self._get_batch(data)
                if not hasattr(self, 'std'):
                    x_rec, z_, mu_ = self.sess.run((self.x_rec_mean, self.z, self.mu), feed_dict={self.x: x_batch})
                else:
                    x_rec, z_, mu_, std_ = self.sess.run((self.x_rec_mean, self.z, self.mu, self.std), feed_dict={self.x: x_batch})
                    zs['std'] = np.vstack((zs['std'], std_))
                zs['map'] = np.vstack((zs['map'], mu_))
                zs['samples'] = np.vstack((zs['samples'], z_))
                if labels is not None:
                    zs['labels'] = np.hstack((zs['labels'], labels[idx]))
        if file_name is not None:
            with open(file_name, 'wb') as postfile:
                pickle.dump(zs, postfile)
        return zs

    def samplePrior(self):
        with self.graph.as_default():
            if isinstance(self.prior, tfd.Dirichlet):
                return self.sess.run(self.x_rec_mean, feed_dict={self.alpha: self.prior_std})
            elif isinstance(self.prior, tfd.MultivariateNormalTriL):
                raise NotImplementedError('Sampling from the MVN prior is not implemented yet.')
            else:
                return self.sess.run(self.x_rec_mean, feed_dict={self.mu: self.prior_mu, self.std: self.prior_std})

    def generate(self, zsamps):
        with self.graph.as_default():
            return self.sess.run(self.x_rec_mean, feed_dict={self.z: zsamps})

    def restore_weights(self, folder_name, file_name=None):
        with self.graph.as_default():
            self.saver.restore(self.sess, folder_name + ((self.name if file_name is None else file_name) + '.ckpt'))

    # ----------------------------------------------------------------------------------- #
    #                                   PLOTTERS                                          #
    # ----------------------------------------------------------------------------------- #

    def plotReconstruction(self, test_data, n_rec, file_name=None, simplified_output=False):
        nRow = 5 if not simplified_output else 2
        f = plt.figure(figsize=(3 * n_rec, 3 * nRow))
        fs = 16
        with self.graph.as_default():
            x_sample, idx = self._get_batch(test_data)
            x_reconstruct, z_vals, z_means = self.sess.run((self.x_rec_mean, self.z, self.mu), feed_dict={self.x: x_sample})
            for i in range(n_rec):
                plt.subplot(nRow, n_rec, i + 1)
                hf.grey_image(x_sample[i], ylabel='input' if i == 0 else None, fontsize=fs, unitscale=isinstance(self.observation, tfd.Bernoulli))
                plt.subplot(nRow, n_rec, n_rec + i + 1)
                hf.grey_image(x_reconstruct[i], ylabel='reconstruction' if i == 0 else None, fontsize=fs,
                              unitscale=simplified_output and isinstance(self.observation, tfd.Bernoulli))
                if not simplified_output:
                    plt.subplot(nRow, n_rec, 2 * n_rec + i + 1)
                    hf.grey_image(x_reconstruct[i], ylabel='reconstr scaled' if i == 0 else None, unitscale=False, fontsize=fs)
                    plt.subplot(nRow, n_rec, 3 * n_rec + i + 1)
                    plt.bar(range(self.latent_dim), z_means[i])
                    if i == 0:
                        plt.ylabel('latent mean', fontsize=fs)
                    plt.subplot(nRow, n_rec, 4 * n_rec + i + 1)
                    plt.bar(range(self.latent_dim), z_vals[i])
                    if i == 0:
                        plt.ylabel('latent sample', fontsize=fs)
        plt.tight_layout()
        if file_name is not None:
            f.savefig(file_name, bbox_inches='tight')

    def plotLatentDistr(self, zs, file_name=None, data='samples'):
        f = plt.figure(figsize=(20, 6))
        plt.subplot(2, 2, 1)
        means = np.mean(zs['map'], axis=0)
        plt.bar(range(self.latent_dim), means)
        plt.subplot(2, 2, 2)
        # stds = np.mean(zs['std'], axis=0)
        stds = np.std(zs['samples'], axis=0)
        plt.bar(range(self.latent_dim), stds)
        plt.subplot(2, 2, 3)
        plt.bar(range(self.latent_dim), np.sort(means))
        plt.subplot(2, 2, 4)
        plt.bar(range(self.latent_dim), np.sort(stds))
        if file_name is not None:
            f.savefig(file_name, bbox_inches='tight')

        if data is not None and data in zs.keys() and 'label' in zs.keys():
            df = pd.DataFrame(zs[data])
            df.columns = ['z%d' % (i + 1) for i in range(self.latent_dim)]
            df['label'] = zs['labels']
            df_long = pd.melt(df, id_vars="label", var_name="latent", value_name="mean")
            sns_plot = sns.catplot("latent", hue="label", y="mean", data=df_long, kind="box", aspect=2, width=0.5)
            if file_name is not None:
                sns_plot.savefig(file_name[0:-4] + '_labelwise.png', bbox_inches='tight')

    def plotReconstrSweep(self, latents, max_val=None, min_val=None, zs=None, file_name=None, grid=True, sweep_samples=None, unitscale=True):
        # latents are indexed from 1 here
        nx = ny = 10
        if max_val is None:
            max_val = zs['samples'].max() if zs is not None else 5
        if min_val is None:
            min_val = zs['samples'].min() if zs is not None else -max_val
        x_values = np.linspace(min_val, max_val, nx)
        y_values = np.linspace(min_val, max_val, ny)
        nPX = int(np.sqrt(self.data_dim))
        canvas = np.empty((nPX * ny, nPX * nx))
        d = np.zeros([self.batch_size, self.latent_dim], dtype='float32')
        d[0, ] = range(1, self.latent_dim + 1)

        xs = np.zeros((nx, ny, self.data_dim))
        for i, yi in enumerate(x_values):
            for j, xi in enumerate(y_values):
                if sweep_samples is None:
                    z_samp = np.zeros((1, self.latent_dim))
                    if zs is not None:
                        z_samp[0, :] = np.mean(zs['samples'], axis=0)
                    z_samp[0, latents[0] - 1], z_samp[0, latents[1] - 1] = xi, yi
                else:
                    z_samp = sweep_samples[i, j, :]
                d[0] = z_samp
                x_mean = self.generate(d)
                xs[i, j, :] = x_mean[0]
                canvas[(nx - i - 1) * nPX:(nx - i) * nPX, j * nPX:(j + 1) * nPX] = x_mean[0].reshape(nPX, nPX)
                if grid:
                    canvas[(nx - i - 1) * nPX:(nx - i) * nPX, (j + 1) * nPX - 1] = 0
                    canvas[(nx - i) * nPX - 1, j * nPX:(j + 1) * nPX] = 0
        f = plt.figure(figsize=(14, 14))
        Xi, Yi = np.meshgrid(x_values, y_values)
        if isinstance(self.observation, tfd.Bernoulli) and unitscale:
            plt.imshow(canvas, origin="upper", vmin=0, vmax=1, interpolation='none', cmap=plt.get_cmap('gray'))
        else:
            plt.imshow(canvas, origin="upper", vmin=xs.min(), vmax=xs.max(), interpolation='none', cmap=plt.get_cmap('gray'))
        plt.gca().set_xticks([])
        plt.gca().set_yticks([])
        plt.xlabel('z%d' % (latents[0]), fontsize=30)
        plt.ylabel('z%d' % (latents[1]), fontsize=30)
        plt.tight_layout()
        if file_name is not None:
            f.savefig(file_name, bbox_inches='tight')
        return xs
    
    @hf.ignore_warnings(PendingDeprecationWarning)  # TSNE uses the matrix class
    def plotLatents(self, zs, latents=None, tsne='samples', file_name=None, cmap='Set1', point_size=1, tsne_perp=20, tsne_ndata=2000):
        # latents are indexed from 1 here
        if latents is None:
            latents = np.argsort(np.abs(np.mean(zs['map'], axis=0)))[-2:] + 1
        f = plt.figure(figsize=(14, 4))
        fs = 12        
        plt.subplot(1, 3, 1)
        plt.scatter(zs['map'][:, latents[0] - 1], zs['map'][:, latents[1] - 1], s=point_size, c=zs["labels"] if 'labels' in zs.keys() else 'b', cmap=cmap)        
        plt.title('Latent means/params', fontsize=fs)
        plt.xlabel('z%d' % (latents[0]), fontsize=fs)
        plt.ylabel('z%d' % (latents[1]), fontsize=fs)
        plt.subplot(1, 3, 2)
        plt.scatter(zs['samples'][:, latents[0] - 1], zs['samples'][:, latents[1] - 1], s=point_size, c=zs["labels"] if 'labels' in zs.keys() else 'b', cmap=cmap)
        plt.title('Latent samples', fontsize=fs)
        plt.xlabel('z%d' % (latents[0]), fontsize=fs)
        plt.ylabel('z%d' % (latents[1]), fontsize=fs)
        if tsne is None:
            plt.subplot(1, 3, 3)
            plt.scatter(zs["labels"], np.zeros(zs["labels"].shape[0]), c=zs["labels"], s=500, cmap=cmap)

        if self.latent_dim > 2 and tsne is not None:
            plt.subplot(1, 3, 3)
            random_indices = np.random.randint(0, zs['map'].shape[0] - 1, size=tsne_ndata)
            model = TSNE(n_components=2, perplexity=tsne_perp, verbose=False)
            np.set_printoptions(suppress=True)
            tsne_data = model.fit_transform(zs[tsne][(random_indices)])
            plt.scatter(tsne_data[:, 0], tsne_data[:, 1], c=np.array(zs["labels"])[(random_indices)] if 'labels' in zs.keys() else 'b', s=15, cmap=cmap)
            plt.title('t-SNE ' + tsne, fontsize=fs)
        if file_name is not None:
            f.savefig(file_name, bbox_inches='tight')
    
    def _latentTraversal(self, latents, rest='random', n_batch=1, max_val=None, zs=None):
        if max_val is None:
            max_val = zs['samples'].max() if zs is not None else 1
            
        rfs = np.zeros((len(latents), self.data_dim))
        with self.graph.as_default():
            for l in range(len(latents)):
                diff = np.zeros((0, self.data_dim))
                for b in range(n_batch):
                    if rest == 'random':
                        if isinstance(self.prior, tfd.MultivariateNormalTriL):
                            raise NotImplementedError('RF calculation with the MVN prior is not implemented yet.')
                        params_feed = {self.alpha: self.prior_std} if isinstance(self.prior, tfd.Dirichlet) else {self.mu: self.prior_mu, self.std: self.prior_std}
                        activations = self.sess.run(self.z, feed_dict=params_feed)
                    else:
                        activations = np.zeros((self.batch_size, self.latent_dim))
                    activations[:, latents[l]] = -max_val
                    x_mean_low = self.generate(activations)
                    activations[:, latents[l]] = max_val
                    x_mean_high = self.generate(activations)
                    diff = np.vstack((diff, x_mean_high - x_mean_low))
                rfs[l, :] = np.mean(diff, axis=0)
            return rfs
        
    def plotLatentTraversalRFs(self, zs=None, file_name=None):
        rfs = self._latentTraversal(range(self.latent_dim), zs=zs)
        hf.image_set(rfs, samplePerRow=20, unitscale=False, file_name=file_name)
    
    def plotReceptiveField(self, latents, max_val=None, rest='random', n_batch=1, zs=None, file_name=None):
        # latents are indexed from 1 here
        if not isinstance(latents, list):
            if zs is not None:
                idx = list(np.argsort(np.abs(np.mean(zs['map'], axis=0)))[-int(latents / 2):]) + list(np.argsort(np.mean(zs['std'], axis=0))[-int(latents / 2):])
            else:
                idx = random.sample(range(self.latent_dim), latents)
            latents = [i + 1 for i in idx]
            print(latents)        
            
        rfs = self._latentTraversal([l - 1 for l in latents], rest, n_batch, max_val, zs)
        
        f = plt.figure(figsize=(3 * len(latents), 6))
        for l in range(len(latents)):
            plt.subplot(2, len(latents), l + 1)
            color = 'red' if l < len(latents) / 2 else 'green'
            hf.grey_image(rfs[l], border_color=color if zs is not None else None, unitscale=1.0 if isinstance(self.observation, tfd.Normal) else True)
            plt.subplot(2, len(latents), len(latents) + l + 1)
            hf.grey_image(rfs[l], unitscale=False)
        if file_name is not None:
            f.savefig(file_name, bbox_inches='tight')
            
    def plotWhiteNoiseRFs(self, n_batches, file_name=None):
        all_img = np.zeros((0, self.data_dim))
        all_latents = np.zeros((0, self.latent_dim))
        with self.graph.as_default():
            for b in range(n_batches):
                img_batch = np.array([hf.get_white_noise_image(self.nPX, self.nPX).flatten() for i in range(self.batch_size)])
                z_ = self.sess.run((self.z), feed_dict={self.x: img_batch})
                all_img = np.vstack([all_img, img_batch])
                all_latents = np.vstack([all_latents, z_])        
        rfs = np.array([[np.corrcoef(all_latents[:, l], all_img[:, px])[0, 1] for px in range(self.data_dim)] for l in range(self.latent_dim)])
        hf.image_set(rfs, samplePerRow=20, unitscale=False, file_name=file_name)
